#!/usr/bin/python

from __future__ import print_function

import aptsources.distro
import aptsources.sourceslist
import unittest
import apt_pkg
import os
import copy

class TestAptSources(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        for k in apt_pkg.config.keys():
            apt_pkg.config.clear(k)
        apt_pkg.init()
		
    def setUp(self):
        self.testdir = os.path.abspath(os.path.dirname(__file__))
        apt_pkg.init()
        apt_pkg.config.set("Dir::Etc", self.testdir)
        apt_pkg.config.set("Dir::Etc::sourceparts",".")
        self.sources_list = os.path.join(self.testdir, "data", "sources.list")
        apt_pkg.config.set("Dir::Etc::sourcelist", self.sources_list)

    def testIsMirror(self):
        self.assertTrue(aptsources.sourceslist.is_mirror("http://archive.ubuntu.com",
                                              "http://de.archive.ubuntu.com"))
        self.assertFalse(aptsources.sourceslist.is_mirror("http://archive.ubuntu.com",
                                              "http://ftp.debian.org"))

    def testSourcesListReading(self):
        apt_pkg.config.set("Dir::Etc::sourcelist", self.sources_list)
        sources = aptsources.sourceslist.SourcesList()
        self.assertEqual(len(sources.list), 6)
        # test load
        sources.list = []
        sources.load(self.sources_list)
        self.assertEqual(len(sources.list), 6)

    def testSourcesListAdding(self):
        apt_pkg.config.set("Dir::Etc::sourcelist", self.sources_list)
        sources = aptsources.sourceslist.SourcesList()
        # test to add something that is already there (main)
        before = copy.deepcopy(sources)
        sources.add("deb","http://de.archive.ubuntu.com/ubuntu/",
                    "edgy",
                    ["main"])
        self.assertTrue(sources.list == before.list)
        # test to add something that is already there (restricted)
        before = copy.deepcopy(sources)
        sources.add("deb","http://de.archive.ubuntu.com/ubuntu/",
                    "edgy",
                    ["restricted"])
        self.assertTrue(sources.list == before.list)
        # test to add something new: multiverse
        sources.add("deb","http://de.archive.ubuntu.com/ubuntu/",
                    "edgy",
                    ["multiverse"])
        found = False
        for entry in sources:
            if (entry.type == "deb" and
                entry.uri == "http://de.archive.ubuntu.com/ubuntu/" and
                entry.dist == "edgy" and
                "multiverse" in entry.comps):
                found = True
        self.assertTrue(found)
        # test to add something new: multiverse *and* 
        # something that is already there
        before = copy.deepcopy(sources)
        sources.add("deb","http://de.archive.ubuntu.com/ubuntu/",
                    "edgy",
                    ["universe", "something"])
        found_universe = 0
        found_something = 0
        for entry in sources:
            if (entry.type == "deb" and
                entry.uri == "http://de.archive.ubuntu.com/ubuntu/" and
                entry.dist == "edgy"):
                for c in entry.comps:
                    if c == "universe":
                        found_universe += 1
                    if c == "something":
                        found_something += 1
        #print("\n".join([s.str() for s in sources]))
        self.assertEqual(found_something, 1)
        self.assertEqual(found_universe, 1)

    def testDistribution(self):
        apt_pkg.config.set("Dir::Etc::sourcelist",
                           self.sources_list+".testDistribution")
        sources = aptsources.sourceslist.SourcesList()
        distro = aptsources.distro.get_distro(codename="edgy", id="Ubuntu", description="Ubuntu Edgy", release="edgy")
        distro.get_sources(sources)
        # test if all suits of the current distro were detected correctly
        dist_templates = set()
        for s in sources:
            if s.template:
                dist_templates.add(s.template.name)
        #print(dist_templates)
        for d in ["edgy","edgy-security","edgy-updates","hoary"]:
            self.assertTrue(d in dist_templates)
        # test enable 
        comp = "restricted"
        distro.enable_component(comp)
        found = {}
        for entry in sources:
            if (entry.type == "deb" and
                entry.uri == "http://de.archive.ubuntu.com/ubuntu/" and
                "edgy" in entry.dist):
                for c in entry.comps:
                    if c == comp:
                        if entry.dist not in found:
                            found[entry.dist] = 0
                        found[entry.dist] += 1
        #print("".join([s.str() for s in sources]))
        for key in found:
            self.assertEqual(found[key], 1)

        # add a not-already available component
        comp = "multiverse"
        distro.enable_component(comp)
        found = {}
        for entry in sources:
            if (entry.type == "deb" and
                entry.template and
                entry.template.name == "edgy"):
                for c in entry.comps:
                    if c == comp:
                        if entry.dist not in found:
                            found[entry.dist] = 0
                        found[entry.dist] += 1
        #print("".join([s.str() for s in sources]))
        for key in found:
            self.assertEqual(found[key], 1)

if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
#
# D-Bus based interface for software-properties
#
# Copyright © 2010 Harald Sitter <apachelogger@ubuntu.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.

from gi.repository import GLib

import dbus.service
import logging
import subprocess
import tempfile
import sys
import threading

from aptsources.sourceslist import SourceEntry

from dbus.mainloop.glib import DBusGMainLoop
from softwareproperties.LivepatchService import LivepatchService
from softwareproperties.SoftwareProperties import SoftwareProperties

DBUS_BUS_NAME = 'com.ubuntu.SoftwareProperties'
DBUS_PATH = '/'
DBUS_INTERFACE_NAME = 'com.ubuntu.SoftwareProperties'

DBusGMainLoop(set_as_default=True)

def _to_unicode(string):
  if sys.version < '3':
    return string.encode('utf-8')
  else:
    return string

class PermissionDeniedByPolicy(dbus.DBusException):
    _dbus_error_name = 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy'

class SoftwarePropertiesDBus(dbus.service.Object, SoftwareProperties):

    def __init__(self, bus, options=None, datadir=None, rootdir="/"):
        # init software properties
        SoftwareProperties.__init__(self, options=options, datadir=datadir, rootdir=rootdir)
        # used in _check_policykit_priviledge
        self.dbus_info = None
        self.polkit = None
        # init dbus service
        bus_name = dbus.service.BusName(DBUS_INTERFACE_NAME, bus=bus)
        dbus.service.Object.__init__(self, bus_name, DBUS_PATH)
        # useful for testing
        self.enforce_polkit = True
        logging.debug("waiting for connections")

        self._livepatch_service = LivepatchService()

    # override set_modified_sourceslist to emit a signal
    def save_sourceslist(self):
        super(SoftwarePropertiesDBus, self).save_sourceslist()
        self.SourcesListModified()
    def write_config(self):
        super(SoftwarePropertiesDBus, self).write_config()
        self.ConfigModified()

    # ------------------ SIGNALS
    @dbus.service.signal(dbus_interface=DBUS_INTERFACE_NAME, signature='')
    def SourcesListModified(self):
        """ emit signal when the sources.list got modified """
        logging.debug("SourcesListModified signal")

    @dbus.service.signal(dbus_interface=DBUS_INTERFACE_NAME, signature='')
    def ConfigModified(self):
        """ emit signal when the sources.list got modified """
        logging.debug("ConfigModified signal")

    @dbus.service.signal(dbus_interface=DBUS_INTERFACE_NAME, signature='')
    def KeysModified(self):
        """ emit signal when the apt keys got modified """
        logging.debug("KeysModified signal")

    @dbus.service.signal(dbus_interface=DBUS_INTERFACE_NAME, signature='')
    def AuthFailed(self):
        """ emit signal when the policykit authentication failed """
        logging.debug("Auth signal")

    @dbus.service.signal(dbus_interface=DBUS_INTERFACE_NAME, signature='')
    def CdromScanFailed(self):
        """ emit signal when adding a cdrom failed """
        logging.debug("Cdrom scan failed signal")

    # ------------------ METHODS

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='', out_signature='')
    def Revert(self, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.revert()

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='', out_signature='')
    def Reload(self, sender=None, conn=None):
        self.reload_sourceslist()

    # Enabler/Disablers
    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='s', out_signature='')
    def EnableChildSource(self, template, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.enable_child_source(_to_unicode(template))

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='s', out_signature='')
    def DisableChildSource(self, template, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.disable_child_source(_to_unicode(template))

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='s', out_signature='')
    def EnableComponent(self, component, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.enable_component(_to_unicode(component))

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='s', out_signature='')
    def DisableComponent(self, component, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.disable_component(_to_unicode(component))

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='', out_signature='')
    def EnableSourceCodeSources(self, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.enable_source_code_sources()

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='', out_signature='')
    def DisableSourceCodeSources(self, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.disable_source_code_sources()
        self.save_sourceslist()

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='s', out_signature='')
    def ToggleSourceUse(self, source, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.toggle_source_use(_to_unicode(source))

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='ss', out_signature='b')
    def ReplaceSourceEntry(self, old, new, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        return self.replace_source_entry(
            _to_unicode(old), _to_unicode(new))

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='', out_signature='')
    def ChangeMainDownloadServer(self, server, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.change_main_download_server(_to_unicode(server))


    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='', out_signature='')
    def AddCdromSource(self, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self._add_cdrom_source()

    def _add_cdrom_source(self):
        """ add a (already inserted) cdrom """
        tmp = tempfile.NamedTemporaryFile()    
        # need to call it here because python-apt does not support
        # AutoDetect mode yet
        cmd = ["apt-cdrom", "add",
               "-o", "Debug::aptcdrom=1",
               "-o", "Debug::identcdrom=1",
               "-o", "acquire::cdrom::AutoDetect=1",
               "-o", "acquire::cdrom::NoMount=1",
               "-o", "Dir::Etc::sourcelist=%s" % tmp.name,
               ]
        p = subprocess.Popen(cmd)
        # wait for the process to finish
        GLib.timeout_add(500, self._wait_for_cdrom_scan_finish, p, tmp)

    def _wait_for_cdrom_scan_finish(self, p, tmp):
        """ glib timeout helper to wait for the cdrom scanner to finish """
        # keep the timeout running
        if p.poll() is None:
            return True
        # else we have a return code
        res = p.poll()
        if res != 0:
            self.CdromScanFailed()
            return False
        # read tmp file with source name 
        line = ""
        # (read only last line)
        for x in open(tmp.name):
            line = x
        if line != "":
            self.sourceslist.list.insert(0, SourceEntry(line))
            self.set_modified_sourceslist()
        return False

    # Setters
    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='b', out_signature='')
    def SetPopconPariticipation(self, participates, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.set_popcon_pariticipation(participates)

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='i', out_signature='')
    def SetUpdateAutomationLevel(self, state, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.set_update_automation_level(state)

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='i', out_signature='')
    def SetReleaseUpgradesPolicy(self, state, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.set_release_upgrades_policy(state)

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='i', out_signature='')
    def SetUpdateInterval(self, days, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.set_update_interval(int(days))

    # Sources

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='s', out_signature='')
    def AddSourceFromLine(self, sourceLine, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.add_source_from_line(_to_unicode(sourceLine))
        self.KeysModified()

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='s', out_signature='')
    def RemoveSource(self, source, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        self.remove_source(_to_unicode(source))

    # GPG Keys

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='s', out_signature='b')
    def AddKey(self, path, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        res = self.add_key(path)
        if res:
            self.KeysModified()
        return res

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='s', out_signature='b')
    def AddKeyFromData(self, keyData, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        res = self.add_key_from_data(keyData)
        if res:
            self.KeysModified()
        return res

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='s', out_signature='b')
    def RemoveKey(self, keyid, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        res = self.remove_key(keyid)
        if res:
            self.KeysModified()
        return res

    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='', out_signature='b')
    def UpdateKeys(self, sender=None, conn=None):
        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        res = self.update_keys()
        if res:
            self.KeysModified()
        return res

    # LivePatch
    @dbus.service.method(DBUS_INTERFACE_NAME,
                         sender_keyword="sender", connection_keyword="conn",
                         in_signature='bs', out_signature='bs', async_callbacks=('reply_handler', 'error_handler'))
    def SetLivepatchEnabled(self, enabled, token, reply_handler, error_handler, sender=None, conn=None):
        def enable_thread_func():
            ret = self._livepatch_service.set_enabled(enabled, token)
            GLib.idle_add(lambda: reply_handler(*ret))

        self._check_policykit_privilege(
            sender, conn, "com.ubuntu.softwareproperties.applychanges")
        threading.Thread(target=enable_thread_func).start()

    # helper from jockey
    def _check_policykit_privilege(self, sender, conn, privilege):
        '''Verify that sender has a given PolicyKit privilege.

        sender is the sender's (private) D-BUS name, such as ":1:42"
        (sender_keyword in @dbus.service.methods). conn is
        the dbus.Connection object (connection_keyword in
        @dbus.service.methods). privilege is the PolicyKit privilege string.

        This method returns if the caller is privileged, and otherwise throws a
        PermissionDeniedByPolicy exception.
        '''
        if sender is None and conn is None:
            # called locally, not through D-BUS
            return
        if not self.enforce_polkit:
            # that happens for testing purposes when running on the session
            # bus, and it does not make sense to restrict operations here
            return

        # get peer PID
        if self.dbus_info is None:
            self.dbus_info = dbus.Interface(conn.get_object('org.freedesktop.DBus',
                '/org/freedesktop/DBus/Bus', False), 'org.freedesktop.DBus')
        pid = self.dbus_info.GetConnectionUnixProcessID(sender)

        # query PolicyKit
        if self.polkit is None:
            self.polkit = dbus.Interface(dbus.SystemBus().get_object(
                'org.freedesktop.PolicyKit1',
                '/org/freedesktop/PolicyKit1/Authority', False),
                'org.freedesktop.PolicyKit1.Authority')
        try:
            # we don't need is_challenge return here, since we call with AllowUserInteraction
            (is_auth, _, details) = self.polkit.CheckAuthorization(
                    ('system-bus-name', {'name': dbus.String(sender, variant_level = 1)}),
                    privilege, {'': ''}, dbus.UInt32(1), '', timeout=600)
        except dbus.DBusException as e:
            if e._dbus_error_name == 'org.freedesktop.DBus.Error.ServiceUnknown':
                # polkitd timed out, connect again
                self.polkit = None
                return self._check_polkit_privilege(sender, conn, privilege)
            else:
                raise

        if not is_auth:
            logging.debug('_check_polkit_privilege: sender %s on connection %s pid %i is not authorized for %s: %s' %
                    (sender, conn, pid, privilege, str(details)))
            self.AuthFailed()
            raise PermissionDeniedByPolicy(privilege)



#  Copyright (c) 2013 Canonical Ltd.
#
#  Author: Scott Moser <smoser@ubuntu.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

import aptsources.distro
from gettext import gettext as _

_DEF_CODENAME = aptsources.distro.get_distro().codename


class ShortcutHandler(object):
    # the defeault ShortcutHandler only handles actual apt lines.
    # ie, 'shortcut' here is a line like you'd find in /etc/apt/sources.list:
    #   deb MIRROR RELEASE-POCKET COMPONENT
    def __init__(self, shortcut):
        self.shortcut = shortcut

    def add_key(self, keyserver=None):
        return True

    def expand(self, codename=None, distro=None):
        return (self.shortcut, None)

    def info(self):
        return {
            'description': _("No description available for '%(shortcut)s'") %
                             {'shortcut': self.shortcut},
            'web_link': _("web link unavailable")}

    def should_confirm(self):
        return False


class ShortcutException(Exception):
    pass


def shortcut_handler(shortcut):
    # this is the default shortcut handler, so it matches anything
    return ShortcutHandler(shortcut)

# vi: ts=4 expandtab



#!/usr/bin/python

import apt_pkg

import os
import unittest
import sys
sys.path.insert(0, "..")

from mock import patch

import softwareproperties.ppa
from  softwareproperties.ppa import (
    AddPPASigningKeyThread,
    mangle_ppa_shortcut,
    verify_keyid_is_v4,
    )


MOCK_PPA_INFO={
    "displayname": "PPA for Michael Vogt",
    "web_link": "https://launchpad.net/~mvo/+archive/ppa", 
    "signing_key_fingerprint": "019A25FED88F961763935D7F129196470EB12F05",
    "name": "ppa",
    'distribution_link': 'https://launchpad.net/api/1.0/ubuntu',
    'owner_link': 'https://launchpad.net/api/1.0/~mvo',
    'reference': '~mvo/ubuntu/ppa',
    'self_link': 'https://launchpad.net/api/devel/~mvo/+archive/ubuntu/ppa',
    }

MOCK_KEY="""
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: SKS 1.1.6
Comment: Hostname: keyserver.ubuntu.com

mI0ESXP67wEEAN2m3xWkAP0p1erHbJx1wYBCL6tLqWXESx1BmF0htLzdD9lfsUYiNs+Zgg3w
uU0PrQIcqZtyTESh514tw3KQ+OAK2I0a2XJR99lXPksiKoxaOOsr0pTVWDYuIlfV3yfmXvnK
FZSmaMjjKuqQbCwZe8Ev7yry9Gh9pM5Y87MbNT05ABEBAAG0HkxhdW5jaHBhZCBQUEEgZm9y
IE1pY2hhZWwgVm9ndIi2BBMBAgAgBQJJc/rvAhsDBgsJCAcDAgQVAggDBBYCAwECHgECF4AA
CgkQEpGWRw6xLwVofAP/YyU3YykXbr8p7wRp1EpFlDmtbPlFXp00gt4Cqlu2AWVOkwkVoMRQ
Ncb7wog2Z6u7KyUhD8pgC2FEL0+FQjyNemv7D0OYBG+6DLdjtRsv0CumLdWFmviU96j3OcwT
G2GkIC/eB2maTrV/vj7vlZ0Qe/T1NL6XLpr0A6Rg6JAtkFM=
=SMbJ
-----END PGP PUBLIC KEY BLOCK-----
"""

MOCK_SECOND_KEY="""
-----BEGIN PGP PUBLIC KEY BLOCK-----
Version: SKS 1.1.6
Comment: Hostname: keyserver.ubuntu.com

mI0ESX34EgEEAOTzplZO3TXmb9dRLu7kOuIEia21e4gwQ/RQe+LD7HdhikcETjf2Ruu0mn6S
sgPLL+duhKxmv6ZciLUgkk0qEDCZuR6BPxdgAIwqmQmFipcv6UTMQitRPUa9WlPU37Qg+joL
cTBUdamnVq+yJhLmnuO44UWAty85nNJzDd29gxqXABEBAAG0LUxhdW5jaHBhZCBQUEEgZm9y
INCU0LzQuNGC0YDQuNC5INCb0LXQtNC60L7Qsoi2BBMBAgAgBQJJffgSAhsDBgsJCAcDAgQV
AggDBBYCAwECHgECF4AACgkQFXlR/kAx0oeuSwQAuhhgWgeeG3F9XMYDqgJShzMSeQOLMKBq
6mNFEL1sDhRdbinf7rwuQFXDSSNCj8/PLa3DF/u09tAm6CTi10iwxxbXf16pTq21gxCA3/xS
fszv352yZpcN85MD5aozqv7qUCGOQ9Gey7JzgD7L4wMEjyRScVjx1chfLgyapdj822E=
=pdql
-----END PGP PUBLIC KEY BLOCK-----
"""

class LaunchpadPPATestCase(unittest.TestCase):
	
    @classmethod
    def setUpClass(cls):
        for k in apt_pkg.config.keys():
            apt_pkg.config.clear(k)
        apt_pkg.init()        

    @unittest.skipUnless(
        "TEST_ONLINE" in os.environ,
        "skipping online tests unless TEST_ONLINE environment variable is set")
    @unittest.skipUnless(
        sys.version_info[0] > 2,
        "pycurl doesn't raise SSL exceptions anymore it seems")
    def test_ppa_info_from_lp(self):
        # use correct data
        info = softwareproperties.ppa.get_ppa_info_from_lp("mvo", "ppa")
        self.assertNotEqual(info, {})
        self.assertEqual(info["name"], "ppa")
        # use empty CERT file
        softwareproperties.ppa.LAUNCHPAD_PPA_CERT = "/dev/null"
        with self.assertRaises(Exception):
            softwareproperties.ppa.get_ppa_info_from_lp("mvo", "ppa")

    def test_mangle_ppa_shortcut(self):
        self.assertEqual("~mvo/ubuntu/ppa", mangle_ppa_shortcut("ppa:mvo"))
        self.assertEqual(
            "~mvo/ubuntu/compiz", mangle_ppa_shortcut("ppa:mvo/compiz"))
        self.assertEqual(
            "~mvo/ubuntu-rtm/compiz",
            mangle_ppa_shortcut("ppa:mvo/ubuntu-rtm/compiz"))

    def test_mangle_ppa_shortcut_leading_slash(self):
        # Test for LP: #1426933
        self.assertEqual("~gottcode/ubuntu/gcppa",
                         mangle_ppa_shortcut("ppa:/gottcode/gcppa"))

    def test_mangle_ppa_supports_no_ppa_colon_prefix(self):
        """mangle_ppa should also support input without 'ppa:'."""
        self.assertEqual("~mvo/ubuntu/ppa", mangle_ppa_shortcut("~mvo/ppa"))


class AddPPASigningKeyTestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        for k in apt_pkg.config.keys():
            apt_pkg.config.clear(k)
        apt_pkg.init()
        cls.trustedgpg = os.path.join(
            os.path.dirname(__file__), "aptroot", "etc", "apt", "trusted.gpg.d")
        try:
            os.makedirs(cls.trustedgpg)
        except:
            pass

    def setUp(self):
        self.t = AddPPASigningKeyThread("~mvo/ubuntu/ppa")

    @patch("softwareproperties.ppa.get_ppa_info_from_lp")
    @patch("softwareproperties.ppa.subprocess")
    def test_fingerprint_len_check(self, mock_subprocess, mock_get_ppa_info):
        """Test that short keyids (<160bit) are rejected""" 
        mock_ppa_info = MOCK_PPA_INFO.copy()
        mock_ppa_info["signing_key_fingerprint"] = "0EB12F05"
        mock_get_ppa_info.return_value = mock_ppa_info
        # do it
        res = self.t.add_ppa_signing_key("~mvo/ubuntu/ppa")
        self.assertFalse(res)
        self.assertFalse(mock_subprocess.Popen.called)
        self.assertFalse(mock_subprocess.call.called)

    @patch("softwareproperties.ppa.get_ppa_info_from_lp")
    @patch("softwareproperties.ppa.get_info_from_https")
    def test_add_ppa_signing_key_wrong_fingerprint(self, mock_https, mock_get_ppa_info):
        mock_get_ppa_info.return_value = MOCK_PPA_INFO
        mock_https.return_value = MOCK_SECOND_KEY
        res = self.t.add_ppa_signing_key("~mvo/ubuntu/ppa")
        self.assertFalse(res)

    @patch("softwareproperties.ppa.get_ppa_info_from_lp")
    @patch("softwareproperties.ppa.get_info_from_https")
    def test_add_ppa_signing_key_multiple_fingerprints(self, mock_https, mock_get_ppa_info):
        mock_get_ppa_info.return_value = MOCK_PPA_INFO
        mock_https.return_value = '\n'.join([MOCK_KEY, MOCK_SECOND_KEY])
        res = self.t.add_ppa_signing_key("~mvo/ubuntu/ppa")
        self.assertFalse(res)

    @patch("softwareproperties.ppa.get_ppa_info_from_lp")
    @patch("softwareproperties.ppa.get_info_from_https")
    @patch("apt_pkg.config")
    def test_add_ppa_signing_key_ok(self, mock_config, mock_https, mock_get_ppa_info):
        mock_get_ppa_info.return_value = MOCK_PPA_INFO
        mock_https.return_value = MOCK_KEY
        mock_config.find_dir.return_value = self.trustedgpg
        res = self.t.add_ppa_signing_key("~mvo/ubuntu/ppa")
        self.assertTrue(res)
    
    def test_verify_keyid_is_v4(self):
        keyid = "0EB12F05"
        self.assertFalse(verify_keyid_is_v4(keyid))


if __name__ == "__main__":
    unittest.main()



#  software-properties cloud-archive support
#
#  Copyright (c) 2013 Canonical Ltd.
#
#  Author: Scott Moser <smoser@ubuntu.org>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

from __future__ import print_function

import apt_pkg
import os
import subprocess
from gettext import gettext as _

from softwareproperties.shortcuts import ShortcutException

RELEASE_MAP = {
    'folsom': 'precise',
    'grizzly': 'precise',
    'havana': 'precise',
    'icehouse': 'precise',
    'juno': 'trusty',
    'kilo': 'trusty',
    'liberty': 'trusty',
    'mitaka': 'trusty',
    'newton': 'xenial',
    'ocata': 'xenial',
    'pike': 'xenial',
    'queens': 'xenial',
    'rocky': 'bionic',
    'stein': 'bionic',
    'train': 'bionic',
    'ussuri': 'bionic',
}
MIRROR = "http://ubuntu-cloud.archive.canonical.com/ubuntu"
UCA = "Ubuntu Cloud Archive"
WEB_LINK = 'https://wiki.ubuntu.com/OpenStack/CloudArchive'
APT_INSTALL_KEY = ['apt-get', '--quiet', '--assume-yes', 'install',
                   'ubuntu-cloud-keyring']

ALIASES = {'tools-updates': 'tools'}
for _r in RELEASE_MAP:
    ALIASES["%s-updates" % _r] = _r

MAP = {
    'tools': {
        'sldfmt': '%(codename)s-updates/cloud-tools',
        'description': UCA + " for cloud-tools (JuJu and MAAS)"},
    'tools-proposed': {
        'sldfmt': '%(codename)s-proposed/cloud-tools',
        'description': UCA + " for cloud-tools (JuJu and MAAS) [proposed]"}
}

for _r in RELEASE_MAP:
    MAP[_r] = {
        'sldfmt': '%(codename)s-updates/' + _r,
        'description': UCA + ' for ' + 'OpenStack ' + _r.capitalize(),
        'release': RELEASE_MAP[_r]}
    MAP[_r + "-proposed"] = {
        'sldfmt': '%(codename)s-proposed/' + _r,
        'description': UCA + ' for ' + 'OpenStack %s [proposed]' % _r.capitalize(),
        'release': RELEASE_MAP[_r]}


class CloudArchiveShortcutHandler(object):
    def __init__(self, shortcut):
        self.shortcut = shortcut

        prefix = "cloud-archive:"

        subs = {'shortcut': shortcut, 'prefix': prefix,
                'ca_names': sorted(MAP.keys())}
        if not shortcut.startswith(prefix):
            raise ValueError(
                _("shortcut '%(shortcut)s' did not start with '%(prefix)s'")
                % subs)

        name_in = shortcut[len(prefix):]
        caname = ALIASES.get(name_in, name_in)

        subs.update({'input_name': name_in})
        if caname not in MAP:
            raise ShortcutException(
                _("'%(input_name)s': not a valid cloud-archive name.\n"
                  "Must be one of %(ca_names)s") % subs)

        self.caname = caname
        self._info = MAP[caname].copy()
        self._info['web_link' ] = WEB_LINK

    def info(self):
        return self._info

    def expand(self, codename, distro=None):
        if codename not in (MAP[self.caname]['release'],
                            os.environ.get("CA_ALLOW_CODENAME")):
            raise ShortcutException(
                _("cloud-archive for %(os_release)s only supported on %(codename)s")
                % {'codename': MAP[self.caname]['release'],
                   'os_release': self.caname.capitalize()})
        dist = MAP[self.caname]['sldfmt'] % {'codename': codename}
        line = ' '.join(('deb', MIRROR, dist, 'main',))
        return (line, _fname_for_caname(self.caname))

    def should_confirm(self):
        return True

    def add_key(self, keyserver=None):
        env = os.environ.copy()
        env['DEBIAN_FRONTEND'] = 'noninteractive'
        try:
            subprocess.check_call(args=APT_INSTALL_KEY, env=env)
        except subprocess.CalledProcessError:
            return False
        return True


def _fname_for_caname(caname):
    # caname is an entry in MAP ('tools' or 'tools-proposed')
    return os.path.join(
        apt_pkg.config.find_dir("Dir::Etc::sourceparts"),
        'cloudarchive-%s.list' % caname)


def shortcut_handler(shortcut):
    try:
        return CloudArchiveShortcutHandler(shortcut)
    except ValueError:
        return None



# dialog_apt_key.py.in - edit the apt keys
#  
#  Copyright (c) 2004 Canonical
#  
#  Author: Michael Vogt <mvo@debian.org>
# 
#  This program is free software; you can redistribute it and/or 
#  modify it under the terms of the GNU General Public License as 
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
# 
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

from __future__ import print_function

import atexit
import datetime
import gettext
import os
import shutil
import subprocess
import tempfile

from subprocess import PIPE

# gettext convenient
_ = gettext.gettext
def dummy(e): return e
N_ = dummy

# some known keys
N_("Ubuntu Archive Automatic Signing Key <ftpmaster@ubuntu.com>")
N_("Ubuntu CD Image Automatic Signing Key <cdimage@ubuntu.com>")
N_("Ubuntu Archive Automatic Signing Key (2012) <ftpmaster@ubuntu.com>")
N_("Ubuntu CD Image Automatic Signing Key (2012) <cdimage@ubuntu.com>")
N_("Ubuntu Extras Archive Automatic Signing Key <ftpmaster@ubuntu.com>")

class AptAuth:
    def __init__(self, rootdir="/"):
        self.rootdir = rootdir
        self.tmpdir = tempfile.mkdtemp()
        self.aptconf = os.path.join(self.tmpdir, 'apt.conf')
        with open(self.aptconf, 'w') as f:
            f.write('DIR "%s";\n' % self.rootdir)
        os.environ['APT_CONFIG'] = self.aptconf
        atexit.register(self._cleanup_tmpdir)

    def _cleanup_tmpdir(self):
        shutil.rmtree(self.tmpdir)

    def list(self):
        cmd = ["/usr/bin/apt-key", "--quiet", "adv", "--with-colons", "--batch", "--fixed-list-mode", "--list-keys"]
        res = []
        p = subprocess.Popen(cmd, stdout=PIPE, stderr=PIPE, universal_newlines=True).stdout
        name = ''
        for line in p:
            fields = line.split(":")
            if fields[0] in ["pub", "uid"]:
                name = fields[9]
            if fields[0] == "pub":
                key = fields[4]
                expiry = datetime.date.fromtimestamp(int(fields[5])).isoformat()
            if not name:
                continue
            res.append("%s %s\n%s" % (key, expiry, _(name)))
            name = ''
        p.close()
        return res

    def add(self, filename):
        cmd = ["/usr/bin/apt-key", "--quiet", "--fakeroot", "add", filename]
        p = subprocess.Popen(cmd, stderr=PIPE)
        return (p.wait() == 0)
        
    def update(self):
        cmd = ["/usr/bin/apt-key", "--quiet", "--fakeroot", "update"]
        p = subprocess.Popen(cmd, stderr=PIPE)
        return (p.wait() == 0)

    def rm(self, key):
        cmd = ["/usr/bin/apt-key", "--quiet", "--fakeroot", "rm", key]
        p = subprocess.Popen(cmd, stderr=PIPE)
        return (p.wait() == 0)



#  software-properties PPA support
#
#  Copyright (c) 2004-2009 Canonical Ltd.
#
#  Author: Michael Vogt <mvo@debian.org>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

from __future__ import print_function

import apt_pkg
import json
import os
import re
import shutil
import subprocess
import tempfile
import time

from gettext import gettext as _
from threading import Thread

from softwareproperties.shortcuts import ShortcutException

try:
    import urllib.request
    from urllib.error import HTTPError, URLError
    import urllib.parse
    from http.client import HTTPException
    NEED_PYCURL = False
except ImportError:
    NEED_PYCURL = True
    import pycurl
    HTTPError = pycurl.error


SKS_KEYSERVER = 'https://keyserver.ubuntu.com/pks/lookup?op=get&options=mr&exact=on&search=0x%s'
# maintained until 2015
LAUNCHPAD_PPA_API = 'https://launchpad.net/api/devel/%s/+archive/%s'
LAUNCHPAD_USER_API = 'https://launchpad.net/api/1.0/%s'
LAUNCHPAD_USER_PPAS_API = 'https://launchpad.net/api/1.0/%s/ppas'
LAUNCHPAD_DISTRIBUTION_API = 'https://launchpad.net/api/1.0/%s'
LAUNCHPAD_DISTRIBUTION_SERIES_API = 'https://launchpad.net/api/1.0/%s/%s'
# Specify to use the system default SSL store; change to a different path
# to test with custom certificates.
LAUNCHPAD_PPA_CERT = "/etc/ssl/certs/ca-certificates.crt"


class CurlCallback:
    def __init__(self):
        self.contents = ''

    def body_callback(self, buf):
        self.contents = self.contents + buf


class PPAException(Exception):

    def __init__(self, value, original_error=None):
        self.value = value
        self.original_error = original_error

    def __str__(self):
        return repr(self.value)


def encode(s):
    return re.sub("[^a-zA-Z0-9_-]", "_", s)


def get_info_from_https(url, accept_json, retry_delays=None):
    """Return the content from url.
    accept_json indicates that:
        a.) Send header Accept: 'application/json'
        b.) Instead of raw content, return json.loads(content)
    retry_delays is None or an iterator (including list or tuple)
        If it is None, no retries will be done.
        If it is an iterator, each value is number of seconds to delay before
        retrying.  For example, retry_delays=(3,5) means to try up to 3
        times, with a 3s delay after first failure and 5s delay after second.
        Retries will not be done on 404."""
    func = _get_https_content_pycurl if NEED_PYCURL else _get_https_content_py3
    data = func(lp_url=url, accept_json=accept_json, retry_delays=retry_delays)
    if accept_json:
        return json.loads(data)
    else:
        return data


def get_info_from_lp(lp_url):
    return get_info_from_https(lp_url, True)

def get_ppa_info_from_lp(owner_name, ppa):
    if owner_name[0] != '~':
        owner_name = '~' + owner_name
    lp_url = LAUNCHPAD_PPA_API % (owner_name, ppa)
    return get_info_from_lp(lp_url)

def series_valid_for_distro(distribution, series):
    lp_url = LAUNCHPAD_DISTRIBUTION_SERIES_API % (distribution, series)
    try:
        get_info_from_lp(lp_url)
        return True
    except PPAException:
        return False

def get_current_series_from_lp(distribution):
    lp_url = LAUNCHPAD_DISTRIBUTION_API % distribution
    return os.path.basename(get_info_from_lp(lp_url)["current_series_link"])


def _get_https_content_py3(lp_url, accept_json, retry_delays=None):
    if retry_delays is None:
        retry_delays = []

    trynum = 0
    err = None
    sleep_waits = iter(retry_delays)
    headers = {"Accept": "application/json"} if accept_json else {}

    while True:
        trynum += 1
        try:
            request = urllib.request.Request(str(lp_url), headers=headers)
            lp_page = urllib.request.urlopen(request,
                                             cafile=LAUNCHPAD_PPA_CERT)
            return lp_page.read().decode("utf-8", "strict")
        except (HTTPException, URLError) as e:
            err = PPAException(
                "Error reading %s (%d tries): %s" % (lp_url, trynum, e.reason),
                e)
            # do not retry on 404. HTTPError is a subclass of URLError.
            if isinstance(e, HTTPError) and e.code == 404:
                break
        try:
            time.sleep(next(sleep_waits))
        except StopIteration:
            break

    raise err


def _get_https_content_pycurl(lp_url, accept_json, retry_delays=None):
    # this is the fallback code for python2
    if retry_delays is None:
        retry_delays = []

    trynum = 0
    sleep_waits = iter(retry_delays)

    while True:
        err_msg = None
        err = None
        trynum += 1
        try:
            callback = CurlCallback()
            curl = pycurl.Curl()
            curl.setopt(pycurl.SSL_VERIFYPEER, 1)
            curl.setopt(pycurl.SSL_VERIFYHOST, 2)
            curl.setopt(pycurl.FOLLOWLOCATION, 1)
            curl.setopt(pycurl.WRITEFUNCTION, callback.body_callback)
            if LAUNCHPAD_PPA_CERT:
                curl.setopt(pycurl.CAINFO, LAUNCHPAD_PPA_CERT)
            curl.setopt(pycurl.URL, str(lp_url))
            if accept_json:
                curl.setopt(pycurl.HTTPHEADER, ["Accept: application/json"])
            curl.perform()
            response = curl.getinfo(curl.RESPONSE_CODE)
            curl.close()

            if response != 200:
                err_msg = "response code %i" % response
        except pycurl.error as e:
            err_msg = str(e)
            err = e

        if err_msg is None:
            return callback.contents

        try:
            time.sleep(next(sleep_waits))
        except StopIteration:
            break

    raise PPAException(
        "Error reading %s (%d tries): %s" % (lp_url, trynum, err_msg),
        original_error=err)


def mangle_ppa_shortcut(shortcut):
    if ":" in shortcut:
        ppa_shortcut = shortcut.split(":")[1]
    else:
        ppa_shortcut = shortcut
    if ppa_shortcut.startswith("/"):
        ppa_shortcut = ppa_shortcut.lstrip("/")
    user = ppa_shortcut.split("/")[0]
    if (user[0] == "~"):
        user = user[1:]
    ppa_path_objs = ppa_shortcut.split("/")[1:]
    ppa_path = []
    if (len(ppa_path_objs) < 1):
        ppa_path = ['ubuntu', 'ppa']
    elif (len(ppa_path_objs) == 1):
        ppa_path.insert(0, "ubuntu")
        ppa_path.extend(ppa_path_objs)
    else:
        ppa_path = ppa_path_objs
    ppa = "~%s/%s" % (user, "/".join(ppa_path))
    return ppa

def verify_keyid_is_v4(signing_key_fingerprint):
    """Verify that the keyid is a v4 fingerprint with at least 160bit"""
    return len(signing_key_fingerprint) >= 160/8


class AddPPASigningKey(object):
    " thread class for adding the signing key in the background "

    def __init__(self, ppa_path, keyserver=None):
        self.ppa_path = ppa_path
        self._homedir = tempfile.mkdtemp()

    def __del__(self):
        shutil.rmtree(self._homedir)

    def gpg_cmd(self, args):
        cmd = "gpg -q --homedir %s --no-default-keyring --no-options --import --import-options %s" % (self._homedir, args)
        return subprocess.Popen(cmd.split(), stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        
    def _recv_key(self, ppa_info):
        signing_key_fingerprint = ppa_info["signing_key_fingerprint"]
        try:
            # double check that the signing key is a v4 fingerprint (160bit)
            if not verify_keyid_is_v4(signing_key_fingerprint):
                print("Error: signing key fingerprint '%s' too short" %
                    signing_key_fingerprint)
                return False
        except TypeError:
            print("Error: signing key fingerprint does not exist")
            return False

        return get_ppa_signing_key_data(ppa_info)

    def _minimize_key(self, key):
        p = self.gpg_cmd("import-minimal,import-export")
        (minimal_key, _) = p.communicate(key.encode())
        
        if p.returncode != 0:
            return False
        return minimal_key

    def _get_fingerprints(self, key):
        fingerprints = []
        p = self.gpg_cmd("show-only --fingerprint --batch --with-colons")
        (output, _) = p.communicate(key)
        if p.returncode == 0:
            for line in output.decode('utf-8').splitlines():
                if line.startswith("fpr:"):
                    fingerprints.append(line.split(":")[9])
        return fingerprints

    def _verify_fingerprint(self, key, expected_fingerprint):
        got_fingerprints = self._get_fingerprints(key)
        if len(got_fingerprints) != 1:
            print("Got '%s' fingerprints, expected only one" %
                  len(got_fingerprints))
            return False
        got_fingerprint = got_fingerprints[0]
        if got_fingerprint != expected_fingerprint:
            print("Fingerprints do not match, not importing: '%s' != '%s'" % (
                    expected_fingerprint, got_fingerprint))
            return False
        return True

    def add_ppa_signing_key(self, ppa_path=None):
        """Query and add the corresponding PPA signing key.

        The signing key fingerprint is obtained from the Launchpad PPA page,
        via a secure channel, so it can be trusted.
        """
        if ppa_path is None:
            ppa_path = self.ppa_path

        try:
            ppa_info = get_ppa_info(mangle_ppa_shortcut(ppa_path))
        except PPAException as e:
            print(e.value)
            return False
        try:
            signing_key_fingerprint = ppa_info["signing_key_fingerprint"]
        except IndexError:
            print("Error: can't find signing_key_fingerprint at %s" % ppa_path)
            return False
        
        #  download the armored_key
        armored_key = self._recv_key(ppa_info)
        if not armored_key:
            return False

        trustedgpgd = apt_pkg.config.find_dir("Dir::Etc::trustedparts")
        apt_keyring = os.path.join(trustedgpgd, encode(ppa_info["reference"][1:]))

        minimal_key = self._minimize_key(armored_key)
        if not minimal_key:
            return False
        
        if not self._verify_fingerprint(minimal_key, signing_key_fingerprint):
            return False

        with open('%s.gpg' % apt_keyring, 'wb') as f:
            f.write(minimal_key)

        return True


class AddPPASigningKeyThread(Thread, AddPPASigningKey):
    # This class is legacy.  There are no users inside the software-properties
    # codebase other than a test case.  It was left in case there were outside
    # users.  Internally, we've changed from having a class implement the
    # tread to explicitly launching a thread and invoking a method in it
    # see check_and_add_key_for_whitelisted_shortcut for how.
    def __init__(self, ppa_path, keyserver=None):
        Thread.__init__(self)
        AddPPASigningKey.__init__(self, ppa_path=ppa_path, keyserver=keyserver)

    def run(self):
        self.add_ppa_signing_key(self.ppa_path)


def _get_suggested_ppa_message(user, ppa_name):
    try:
        msg = []
        try:
            try:
                lp_user = get_info_from_lp(LAUNCHPAD_USER_API % user)
            except PPAException:
                return _("ERROR: '{user}' user or team does not exist.").format(user=user)
            lp_ppas = get_info_from_lp(LAUNCHPAD_USER_PPAS_API % user)
            entity_name = _("team") if lp_user["is_team"] else _("user")
            if lp_ppas["total_size"] > 0:
                # Translators: %(entity)s is either "team" or "user"
                msg.append(_("The %(entity)s named '%(user)s' has no PPA named '%(ppa)s'") % {
                        'entity' : entity_name,
                         'user' : user,
                         'ppa' : ppa_name})
                msg.append(_("Please choose from the following available PPAs:"))
                for ppa in lp_ppas["entries"]:
                    msg.append(_(" * '%(name)s':  %(displayname)s") % {
                                 'name' : ppa["name"],
                                 'displayname' : ppa["displayname"]})
            else:
                # Translators: %(entity)s is either "team" or "user"
                msg.append(_("The %(entity)s named '%(user)s' does not have any PPA") % {
                             'entity' : entity_name, 'user' : user})
            return '\n'.join(msg)
        except KeyError:
            return ''
    except ImportError:
        return _("Please check that the PPA name or format is correct.")


def get_ppa_info(shortcut):
    user = shortcut.split("/")[0]
    ppa = "/".join(shortcut.split("/")[1:])
    try:
        ret = get_ppa_info_from_lp(user, ppa)
        ret["distribution"] = ret["distribution_link"].split('/')[-1]
        ret["owner"] = ret["owner_link"].split('/')[-1]
        return ret
    except (HTTPError, Exception):
        msg = []
        msg.append(_("Cannot add PPA: 'ppa:%s/%s'.") % (
            user, ppa))

        # If the PPA does not exist, then try to find if the user/team
        # exists. If it exists, list down the PPAs
        raise ShortcutException('\n'.join(msg) + "\n" +
                                _get_suggested_ppa_message(user, ppa))

    except (ValueError, PPAException):
        raise ShortcutException(
            _("Cannot access PPA (%s) to get PPA information, "
              "please check your internet connection.") % \
              (LAUNCHPAD_PPA_API % (user, ppa)))


def get_ppa_signing_key_data(info=None):
    """Return signing key data in armored ascii format for the provided ppa.
    
    If 'info' is a dictionary, it is assumed to be the result
    of 'get_ppa_info(ppa)'.  If it is a string, it is assumed to
    be a ppa_path.

    Return value is a text string."""
    if isinstance(info, dict):
        link = info["self_link"]
    else:
        link = get_ppa_info(mangle_ppa_shortcut(info))["self_link"]

    return get_info_from_https(link + "?ws.op=getSigningKeyData",
                               accept_json=True, retry_delays=(1, 2, 3))


class PPAShortcutHandler(object):
    def __init__(self, shortcut):
        super(PPAShortcutHandler, self).__init__()
        try:
            self.shortcut = mangle_ppa_shortcut(shortcut)
        except:
            raise ShortcutException(_("ERROR: '{shortcut}' is not a valid ppa format")
                                      .format(shortcut=shortcut))
        info = get_ppa_info(self.shortcut)

        if "private" in info and info["private"]:
            raise ShortcutException(
                _("Adding private PPAs is not supported currently"))

        self._info = info

    def info(self):
        return self._info

    def expand(self, codename, distro=None):
        if (distro is not None
                and distro != self._info["distribution"]
                and not series_valid_for_distro(self._info["distribution"], codename)):
            # The requested PPA is for a foreign distribution.  Guess that
            # the user wants that distribution's current series.
            # This only applies if the local distribution is not the same
            # distribution the remote PPA is associated with AND the local
            # codename is not equal to the PPA's series.
            # e.g. local:Foobar/xenial and ppa:Ubuntu/xenial will use 'xenial'
            #      local:Foobar/fluffy and ppa:Ubuntu/xenial will use '$latest'
            codename = get_current_series_from_lp(self._info["distribution"])
        debline = "deb http://ppa.launchpad.net/%s/%s/%s %s main" % (
            self._info["owner"][1:], self._info["name"],
            self._info["distribution"], codename)
        sourceslistd = apt_pkg.config.find_dir("Dir::Etc::sourceparts")
        filename = os.path.join(sourceslistd, "%s-%s-%s-%s.list" % (
            encode(self._info["owner"][1:]), encode(self._info["distribution"]),
            encode(self._info["name"]), codename))
        return (debline, filename)

    def should_confirm(self):
        return True

    def add_key(self, keyserver=None):
        apsk = AddPPASigningKey(self._info["reference"], keyserver=keyserver)
        return apsk.add_ppa_signing_key()


def shortcut_handler(shortcut):
    if not shortcut.startswith("ppa:"):
        return None
    return PPAShortcutHandler(shortcut)


if __name__ == "__main__":
    import sys
    ppa = mangle_ppa_shortcut(sys.argv[1])
    print(get_ppa_info(ppa))



#
#  Copyright (c) 2019 Canonical
#
#  Authors:
#       Andrea Azzarone <andrea.azzarone@canonical.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

from gettext import gettext as _
import logging

import gi
from gi.repository import Gio, GLib

try:
    gi.require_version('Snapd', '1')
    from gi.repository import Snapd
except(ImportError, ValueError):
    pass


class LivepatchSnap(object):

    # Constants
    SNAP_NAME = 'canonical-livepatch'

    # Public API
    def __init__(self):
        self._snapd_client = Snapd.Client()
        self._cancellable = Gio.Cancellable()

    def get_status(self):
        """ Get the status of canonical-livepatch snap.

        Returns:
            Snapd.SnapStatus.Enun: An enum indicating the status of the snap.
        """
        snap = self._get_raw_snap()
        return snap.get_status() if snap else Snapd.SnapStatus.UNKNOWN

    def enable_or_install(self):
        """Enable or install canonical-livepatch snap.

        Returns:
            (True, '') if successful, (False, error_message) otherwise.
        """
        status = self.get_status()

        if status == Snapd.SnapStatus.ACTIVE:
            logging.warning('{} snap is already active'.format(self.SNAP_NAME))
            return True, ''
        elif status == Snapd.SnapStatus.AVAILABLE:
            return self._install()
        elif status == Snapd.SnapStatus.INSTALLED:
            return self._enable()
        else:
            logging.warning('{} snap is in an unknown state'.format(self.SNAP_NAME))
            return False, _('Canonical Livepatch snap cannot be installed.')

    # Private methods
    def _get_raw_snap(self):
        """Get the Sanpd.Snap raw object of the canonical-livepatch snapd.

        Returns:
            Sanpd.Snap if successful, None otherwise.
        """
        try:
            snap = self._snapd_client.get_snap_sync(
                name=self.SNAP_NAME,
                cancellable=self._cancellable)
        except GLib.Error as e:
            logging.debug('Snapd.Client.get_snap_sync failed: {}'.format(e.message))
            snap = None

        if snap:
            return snap

        try:
            (snaps, ignored) = self._snapd_client.find_sync(
                flags=Snapd.FindFlags.MATCH_NAME,
                query=self.SNAP_NAME,
                cancellable=self._cancellable)
            snap = snaps[0]
        except GLib.Error as e:
            logging.debug('Snapd.Client.find_sync failed: {}'.format(e.message))

        return snap

    def _install(self):
        """Install canonical-livepatch snap.

        Returns:
            (True, '') if successful, (False, error_message) otherwise.
        """
        assert self.get_status() == Snapd.SnapStatus.AVAILABLE

        try:
            self._snapd_client.install2_sync(
                flags=Snapd.InstallFlags.NONE,
                name=self.SNAP_NAME,
                cancellable=self._cancellable)
        except GLib.Error as e:
            return False, _('Canonical Livepatch snap cannot be installed: {}'.format(e.message))
        else:
            return True, ''

    def _enable(self):
        """Enable the canonical-livepatch snap.

        Returns:
            (True, '') if successful, (False, error_message) otherwise.
        """
        assert self.get_status() == Snapd.SnapStatus.INSTALLED

        try:
            self._snapd_client.enable_sync(
                name=self.SNAP_NAME,
                cancellable=self._cancellable)
        except GLib.Error as e:
            return False, _('Canonical Livepatch snap cannot be enabled: {}'.format(e.message))
        else:
            return True, ''



#
#  Copyright (c) 2017-2019 Canonical
#
#  Authors:
#       Andrea Azzarone <andrea.azzarone@canonical.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

import os

from gettext import gettext as _

from softwareproperties.gtk.utils import (
    setup_ui,
)


class DialogLivepatchError:

    RESPONSE_SETTINGS = 100
    RESPONSE_IGNORE = 101

    primary = _("Sorry, there's been a problem with setting up Canonical Livepatch.")

    def __init__(self, parent, datadir):
        """setup up the gtk dialog"""
        self.parent = parent

        setup_ui(
            self,
            os.path.join(datadir, "gtkbuilder", "dialog-livepatch-error.ui"),
            domain="software-properties")

        self.dialog = self.messagedialog_livepatch
        self.dialog.set_transient_for(parent)

    def run(self, error, show_settings_button):
        p = "<span weight=\"bold\" size=\"larger\">{}</span>".format(self.primary)
        self.label_primary.set_markup(p)

        textbuffer = self.treeview_message.get_buffer()
        textbuffer.set_text(error)

        self.button_settings.set_visible(show_settings_button)
        res = self.dialog.run()
        self.dialog.hide()
        return res

    def on_button_settings_clicked(self, b, d=None):
        self.dialog.response(self.RESPONSE_SETTINGS)

    def on_button_ignore_clicked(self, b, d=None):
        self.dialog.response(self.RESPONSE_IGNORE)



#
#  Copyright (c) 2019 Canonical
#
#  Authors:
#       Andrea Azzarone <andrea.azzarone@canonical.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

import datetime
import gettext
from gettext import gettext as _
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gio, GLib, GObject, Gtk
import logging

from softwareproperties.GoaAuth import GoaAuth
from softwareproperties.LivepatchService import (
    LivepatchService,
    LivepatchAvailability)
from .DialogAuth import DialogAuth
from .DialogLivepatchError import DialogLivepatchError


class LivepatchPage(object):

    # Constants
    COMMON_ISSUE_URL = 'https://wiki.ubuntu.com/Kernel/Livepatch#CommonIssues'
    GENERIC_ERR_MSG = _('Canonical Livepatch has experienced an internal error.'
        ' Please refer to {} for further information.'.format(COMMON_ISSUE_URL))

    def __init__(self, parent):
        self._parent = parent

        self._timeout_handler = -1
        self._waiting_livepatch_response = False

        self._lps = LivepatchService()
        self._auth = GoaAuth()

        self._un_settings = None
        source = Gio.SettingsSchemaSource.get_default()
        if source is not None:
            schema = source.lookup('com.ubuntu.update-notifier', True)
            if schema is not None:
                settings = Gio.Settings.new('com.ubuntu.update-notifier')
                if schema.has_key('show-livepatch-status-icon'):
                    self._un_settings = settings

        # Connect signals
        self._lps.connect(
            'notify::availability', self._lps_availability_changed_cb)
        self._lps.connect(
            'notify::enabled', self._lps_enabled_changed_cb)
        self._auth.connect(
            'notify', self._auth_changed_cb)
        self._state_set_handler = self._parent.switch_livepatch.connect(
            'state-set', self._switch_state_set_cb)
        self._parent.button_livepatch_login.connect(
            'clicked', self._button_livepatch_login_clicked_cb)
        self._parent.checkbutton_livepatch_topbar.connect(
            'toggled', self._checkbutton_livepatch_topbar_toggled_cb)

        if self._un_settings is not None:
            self._un_settings_handler = self._un_settings.connect(
                'changed::show-livepatch-status-icon', self._un_settings_changed_cb)

        self._lps.trigger_availability_check()

    @property
    def waiting_livepatch_response(self):
        return self._waiting_livepatch_response

    # Private methods
    def _trigger_ui_update(self, skip=False, error_message=None):
        """Trigger the update of every single user interface component according
        to the current state.

        Args:
            skip (bool): whether to trigger the update after a small timeout.
                Defaults to False.
            error_message (str): error message to display. Defaults to None.
        """
        def do_ui_update():
            self._timeout_handler = -1

            self._update_switch()
            self._update_spinner()
            self._update_switch_label()
            self._update_auth_button()
            self._update_stack(error_message)
            self._update_topbar_checkbutton()

            return False

        if self._timeout_handler > 0:
            GObject.source_remove(self._timeout_handler)
            self._timeout_handler = -1

        if skip:
            do_ui_update()
        else:
            self._timeout_handler = GLib.timeout_add_seconds(2, do_ui_update)

    def _update_switch(self):
        """Update the state of the on/off switch."""
        switch = self._parent.switch_livepatch

        availability = self._lps.props.availability
        enabled = self._lps.props.enabled
        logged = self._auth.logged

        switch.set_sensitive(
            availability == LivepatchAvailability.TRUE and
            (enabled or logged))

        if self._waiting_livepatch_response:
            return

        self._parent.switch_livepatch.handler_block(self._state_set_handler)
        switch.set_state(switch.get_sensitive() and enabled)
        self._parent.switch_livepatch.handler_unblock(self._state_set_handler)

    def _update_spinner(self):
        """Update the state of the in-progress spinner."""
        spinner = self._parent.spinner_livepatch
        availability = self._lps.props.availability

        spinner.set_visible(availability == LivepatchAvailability.CHECKING)
        spinner.props.active = (availability == LivepatchAvailability.CHECKING)

    def _update_switch_label(self):
        """Update the text of the label next to the on/off switch."""
        availability = self._lps.props.availability
        logged = self._auth.logged

        if availability == LivepatchAvailability.CHECKING:
            msg = _('Checking availability…')
        elif availability == LivepatchAvailability.NO_CONNECTIVITY:
            msg = _('Livepatch requires an Internet connection.')
        elif availability == LivepatchAvailability.FALSE:
            msg = _('Livepatch is not available for this system.')
        else:
            if self._parent.switch_livepatch.get_active():
                msg = _("Livepatch is on.")
            elif not logged:
                msg = _("To use Livepatch you need to sign in.")
            else:
                msg = _("Livepatch is off.")

        self._parent.label_livepatch_switch.set_label(msg)

    def _update_auth_button(self):
        """Update the state and the label of the authentication button."""
        button = self._parent.button_livepatch_login

        availability = self._lps.props.availability
        logged = self._auth.logged

        button.set_visible(
            availability == LivepatchAvailability.TRUE and
            not self._parent.switch_livepatch.get_active())
        button.set_label(_('Sign Out') if logged else _('Sign In…'))

    def _update_stack(self, error_message):
        """Update the state of the stack.

        If livepatch is not available nothing will be shown, if an error
        occurred an error message will be shown in a text view, otherwise the
        current livepatch status (e.g. a list of CVE fixes) will be shown.

        Args:
            error_message (str): error message to display.
        """
        availability = self._lps.props.availability
        availability_message = self._lps.props.availability_message

        has_error = (
            error_message is not None or
            (availability == LivepatchAvailability.FALSE and
             availability_message is not None))

        if has_error:
            self._parent.stack_livepatch.set_visible_child_name('page_livepatch_message')
            self._parent.stack_livepatch.set_visible(True)
            text_buffer = self._parent.textview_livepatch.get_buffer()
            text_buffer.delete(
                text_buffer.get_start_iter(), text_buffer.get_end_iter())
            text_buffer.insert_markup(
                text_buffer.get_end_iter(),
                error_message or availability_message, -1)
            return

        if availability == LivepatchAvailability.CHECKING or not self._parent.switch_livepatch.get_active():
            self._parent.stack_livepatch.set_visible(False)
        else:
            self._update_status()

    def _update_topbar_checkbutton(self):
        """Update the state of the checkbutton to show/hide Livepatch status in
        the top bar.
        """
        visibile = self._un_settings is not None
        self._parent.checkbutton_livepatch_topbar.set_visible(visibile)

        if visibile:
            availability = self._lps.props.availability

            self._parent.checkbutton_livepatch_topbar.set_sensitive(
                availability == LivepatchAvailability.TRUE and
                self._lps.props.enabled)
            self._parent.checkbutton_livepatch_topbar.set_active(
                availability == LivepatchAvailability.TRUE and
                self._lps.props.enabled and
                self._un_settings.get_boolean('show-livepatch-status-icon'))

    def _format_timedelta(self, td):
        days = td.days
        hours = td.seconds // 3600
        minutes = td.seconds // 60

        if days > 0:
            return gettext.ngettext(
                    '({} day ago)',
                    '({} days ago)',
                    days).format(days)
        elif hours > 0:
            return gettext.ngettext(
                    '({} hour ago)',
                    '({} hours ago)',
                    hours).format(hours)
        elif minutes > 0:
            return gettext.ngettext(
                    '({} minute ago)',
                    '({} minutes ago)',
                    minutes).format(minutes)
        else:
            return ''

    def _datetime_to_str(self, dt):
        gdt = GLib.DateTime.new_from_unix_utc(dt.timestamp())
        td = datetime.datetime.now(dt.tzinfo) - dt
        return '{} {}'.format(
            gdt.to_local().format('%x %H:%M'),
            self._format_timedelta(td))

    def _update_status(self):
        """Populate the UI to reflect the Livepatch status"""
        status = self._lps.get_status()

        if status is None:
            if not self._waiting_livepatch_response:
                self._trigger_ui_update(skip=True, error_message=_('Failed to retrieve Livepatch status.'))
                return

        self._parent.stack_livepatch.set_visible_child_name('page_livepatch_status')
        self._parent.stack_livepatch.set_visible(True)

        check_state = status['Status'][0]['Livepatch']['CheckState'] if status else None
        state = status['Status'][0]['Livepatch']['State'] if status else None

        if check_state == 'check-failed':
            self._trigger_ui_update(skip=True, error_message=self.GENERIC_ERR_MSG)
            return

        if state in ['applied-with-bug', 'apply-failed', 'unknown']:
            self._trigger_ui_update(skip=True, error_message=self.GENERIC_ERR_MSG)
            return

        self._parent.label_livepatch_last_update.set_label(
                _("Last check for updates: {}").format(
                    self._datetime_to_str(status['Last-Check']) if status else _('None yet')))

        if state in ['unapplied', 'nothing-to-apply'] or state == None:
            self._parent.label_livepatch_header.set_label(_('No updates currently applied.'))
            self._parent.scrolledwindow_livepatch_fixes.set_visible(False)
        elif state == 'applied':
            self._parent.label_livepatch_header.set_label(_('Updates currently applied:'))
            self._parent.scrolledwindow_livepatch_fixes.set_visible(True)
            self._update_fixes(status['Status'][0]['Livepatch']['Fixes'])
        else:
            logging.warning('Livepatch status contains an invalid state: {}'.format(state))

        if check_state == 'needs-check' or state == 'unapplied' or state == 'applying':
            self._trigger_ui_update()

    def _update_fixes(self, fixes):
        """Populate the UI to show the list of applied CVE fixes."""
        treeview = self._parent.treeview_livepatch
        liststore = treeview.get_model()
        liststore.clear()

        for fix in fixes:
            fix_iter = liststore.append()
            liststore.set(fix_iter, [0], [self._format_fix(fix)])

    def _format_fix(self, fix):
        """Format a fix in a UI friendly text."""
        return '<b>{}</b>\n{}'.format(
            fix['Name'], fix['Description'].replace('\n', ' '))

    def _do_login(self):
        """Start the authentication flow to retrieve the livepatch token."""
        dialog = DialogAuth(self._parent.window_main, self._parent.datadir)

        if dialog.run() == Gtk.ResponseType.OK:
            self._auth.login(dialog.account)
            self._parent.switch_livepatch.set_state(True)

    def _do_logout(self):
        """Start the de-authentication flow."""
        self._auth.logout()

    # Signals handler
    def _lps_availability_changed_cb(self, o, v):
        self._trigger_ui_update(skip=True)

    def _lps_enabled_changed_cb(self, o, v):
        if self._waiting_livepatch_response:
            return
        self._trigger_ui_update(skip=False)

    def _auth_changed_cb(self, o, v):
        self._trigger_ui_update(skip=True)

    def _switch_state_set_cb(self, widget, state):
        if not self._waiting_livepatch_response:
            self._waiting_livepatch_response = True

            token = self._auth.token or ''
            self._parent.backend.SetLivepatchEnabled(
                state, token,
                reply_handler=self._enabled_reply_handler,
                error_handler=self._enabled_error_handler,
                timeout=1200)

        self._trigger_ui_update(skip=True)
        self._parent.switch_livepatch.set_state(state)

        return False

    def _button_livepatch_login_clicked_cb(self, button):
        if self._auth.logged:
            self._do_logout()
        else:
            self._do_login()

    def _checkbutton_livepatch_topbar_toggled_cb(self, button):
        if not button.get_sensitive():
            return
        self._un_settings.handler_block(self._un_settings_handler)
        self._un_settings.set_boolean('show-livepatch-status-icon', button.get_active())
        self._un_settings.handler_unblock(self._un_settings_handler)


    def _un_settings_changed_cb(self, settings, key):
        self._trigger_ui_update(skip=True)

    def _show_error_dialog(self, message):
        dialog = DialogLivepatchError(
            self._parent.window_main,
            self._parent.datadir)

        response = dialog.run(
            error=message,
            show_settings_button=not self._parent.window_main.is_visible())

        if response == DialogLivepatchError.RESPONSE_SETTINGS:
            self._parent.window_main.show()
            self._parent.notebook_main.set_current_page(6)
        elif not self._parent.window_main.is_visible():
            self._parent.on_close_button(None)

    # DBus replay handlers
    def _enabled_reply_handler(self, is_error, prompt):
        if self._parent.switch_livepatch.get_active() == self._lps.props.enabled:
            self._waiting_livepatch_response = False
            self._trigger_ui_update(skip=True)

            if not self._parent.window_main.is_visible():
                self._parent.on_close_button(None)
        else:
            if is_error:
                self._waiting_livepatch_response = False
                self._trigger_ui_update(skip=True)
                self._show_error_dialog(prompt)
            else:
                # The user tooggled on/off the switch while we were waiting
                # livepatch to respond back.
                token = self._auth.token or ''
                self._parent.backend.SetLivepatchEnabled(
                    self._parent.switch_livepatch.get_active(),
                    token,
                    reply_handler=self._enabled_reply_handler,
                    error_handler=self._enabled_error_handler,
                    timeout=1200)

    def _enabled_error_handler(self, e):
        self._waiting_livepatch_response = False
        self._trigger_ui_update(skip=True)

        if e._dbus_error_name == 'com.ubuntu.SoftwareProperties.PermissionDeniedByPolicy':
            logging.warning("Authentication canceled, changes have not been saved")

            if not self._parent.window_main.is_visible():
                self._parent.on_close_button(None)
        else:
            self._show_error_dialog(str(e))



#
#  Copyright (c) 2018 Canonical
#
#  Authors:
#       Andrea Azzarone <andrea.azzarone@canonical.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

from enum import IntEnum
import os

from gettext import gettext as _
from softwareproperties.gtk.utils import (
    setup_ui,
)

import gi
gi.require_version('Goa', '1.0')
from gi.repository import Gio, GLib, Goa, Gtk
import logging

class Column(IntEnum):
    ID = 0
    MAIL = 1
    ACCOUNT = 2

class DialogAuth:

    def __init__(self, parent, datadir):
        """setup up the gtk dialog"""
        self.parent = parent

        setup_ui(self, os.path.join(datadir, "gtkbuilder", "dialog-auth.ui"),
            domain="software-properties")

        self.dialog = self.dialog_auth
        self.dialog.set_title('')
        self.dialog.set_deletable(False)
        self.dialog.set_transient_for(parent)

        self.button_continue.grab_focus()

        self.account = None
        self.dispose_on_new_account = False
        self.goa_client = Goa.Client.new_sync(None)

        self._setup_model()
        self._check_ui(select=False)

        # Be ready to other accounts
        self.goa_client.connect('account-added', self._account_added_cb)
        self.goa_client.connect('account-removed', self._account_removed_cb)

    def run(self):
        res = self.dialog.run()
        self.dialog.hide()
        return res

    def _setup_model(self):
        for obj in self.goa_client.get_accounts():
            self._add_account(obj.get_account(), select=False)

    def _set_header(self, label):
        self.label_header.set_markup(
            "<span size='larger' weight='bold'>%s</span>" % label)

    def _check_ui(self, select):
        naccounts = len(self.liststore_account)

        if naccounts == 0:
            self._set_header(
                _('To use Livepatch, you need to use an Ubuntu One account.'))
            self.combobox_account.set_visible(False)
            self.label_account.set_visible(False)
            self.button_add_another.set_visible(False)
            self.button_continue.set_label(_('Sign In / Register…'))
        elif naccounts == 1:
            self._set_header(
                _('To use Livepatch, you need to use your Ubuntu One account.'))
            self.combobox_account.set_visible(False)
            self.label_account.set_visible(True)
            self.label_account.set_text(self.liststore_account[0][Column.MAIL])
            self.button_add_another.set_visible(True)
            self.button_continue.set_label(_('Continue'))
        else:
            self._set_header(
                _('To use Livepatch, you need to use an Ubuntu One account.'))
            self.button_add_another.set_visible(True)
            self.combobox_account.set_visible(True)
            self.label_account.set_visible(False)
            self.button_continue.set_label(_('Use'))
            if select:
                self.combobox_account.set_active(naccounts-1)
            elif self.combobox_account.get_active() == -1:
                self.combobox_account.set_active(0)

    def _ignore_account(self, account):
        return account.props.provider_type != 'ubuntusso'

    def _get_account_iter(self, account):
        row = self.liststore_account.get_iter_first()
        while row is not None:
            account_id =  self.liststore_account.get_value(row, Column.ID)
            if account_id == account.props.id:
                return row
            row = self.liststore_account.iter_next(row)
        return None

    def _add_account(self, account, select):
        if self._ignore_account(account):
            return

        account_iter = self._get_account_iter(account)
        if account_iter is not None:
            return

        account_iter = self.liststore_account.append()
        self.liststore_account.set(account_iter,
            [Column.ID, Column.MAIL, Column.ACCOUNT],
            [account.props.id, account.props.presentation_identity, account])
        self._check_ui(select)

    def _remove_account(self, account):
        if self._ignore_account(account):
            return

        account_iter = self._get_account_iter(account)
        if account_iter is None:
            return

        self.liststore_account.remove(account_iter)
        self._check_ui(select=False)

    def _response_if_valid(self, account):
        def cb(source, res, data):
            try:
                source.call_ensure_credentials_finish(res)
                valid = True
            except GLib.Error as e:
                logging.warning("call_ensure_credentials_finish exception: %s",
                   e.message)
                valid = False
            
            if not valid:
                try:
                    self._spawn_goa_with_args(account.props.id, None)
                except GLib.Error as e:
                    logging.warning ('Failed to spawn gnome-control-center: %s',
                        e.message)
            else:
                self.account = account
                self.dialog.response(Gtk.ResponseType.OK)

        account.call_ensure_credentials(None, cb, None)

    def _build_dbus_params(self, action, arg):
        builder = GLib.VariantBuilder.new(GLib.VariantType.new('av'))

        if action is None and arg is None:
            s = GLib.Variant.new_string('')
            v = GLib.Variant.new_variant(s)
            builder.add_value(v)
        else:
            if action is not None:
                s = GLib.Variant.new_string(action)
                v = GLib.Variant.new_variant(s)
                builder.add_value(v)
            if arg is not None:
                s = GLib.Variant.new_string(arg)
                v = GLib.Variant.new_variant(s)
                builder.add_value(v)

        array = GLib.Variant.new_tuple(
            GLib.Variant.new_string('online-accounts'), builder.end())
        array = GLib.Variant.new_variant(array)

        param = GLib.Variant.new_tuple(
            GLib.Variant.new_string('launch-panel'),
            GLib.Variant.new_array(GLib.VariantType.new('v'), [array]),
            GLib.Variant.new_array(GLib.VariantType.new('{sv}'), None))
        return param

    def _spawn_goa_with_args(self, action, arg):
        proxy = Gio.DBusProxy.new_for_bus_sync(Gio.BusType.SESSION,
            Gio.DBusProxyFlags.NONE, None,
            'org.gnome.ControlCenter',
            '/org/gnome/ControlCenter',
            'org.gtk.Actions', None)

        param = self._build_dbus_params(action, arg)
        proxy.call_sync('Activate', param, Gio.DBusCallFlags.NONE, -1, None)

    def _account_added_cb(self, goa_client, goa_object):
        account = goa_object.get_account()
        if self._ignore_account(account):
            return
        if not self.dispose_on_new_account:
            self._add_account(account, True)
        else:
            self._response_if_valid(account)

    def _account_removed_cb(self, goa_client, goa_object):
        account = goa_object.get_account()
        if not self._ignore_account(account):
            self._remove_account(account)

    def _button_add_another_clicked_cb(self, button):
        try:
            # There is no easy way to put this to false if the user close the
            # windows without adding an account.
            self._spawn_goa_with_args('add', 'ubuntusso')
            self.dispose_on_new_account = True
        except GLib.Error as e:
            logging.warning ('Failed to spawn control-center: %s', e.message)

    def _button_cancel_clicked_cb(self, button):
        self.dialog.response(Gtk.ResponseType.CANCEL)

    def _button_continue_clicked_cb(self, button):
        naccounts = len(self.liststore_account)

        account = None
        if naccounts >= 1:
            active_index = self.combobox_account.get_active()
            account = self.liststore_account[active_index][Column.ACCOUNT]

        if account is None:
            self._button_add_another_clicked_cb(self.button_add_another)
        else:
            self._response_if_valid(account)



#  __init__.py - defines globals
#
#  Copyright (c) 2007 Canonical Ltd.
#
#  Author: Sebastian Heinlein <glatzor@ubuntu.com>
#          Michael Vogt <mvo@debian.org>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

# Level of the update automation
(UPDATE_MANUAL, 
 UPDATE_NOTIFY, 
 UPDATE_DOWNLOAD, 
 UPDATE_INST_SEC) = list(range(4))

# Provide more readable synonyms for apt configuration options
CONF_MAP = {
    "autoupdate"   : "APT::Periodic::Update-Package-Lists",
    "autodownload" : "APT::Periodic::Download-Upgradeable-Packages",
    "autoclean"    : "APT::Periodic::AutocleanInterval",
    "unattended"   : "APT::Periodic::Unattended-Upgrade",
    "max_size"     : "APT::Archives::MaxSize",
    "max_age"      : "APT::Archives::MaxAge"
}





# Copyright (C) 2009 Canonical
#
# Authors:
#  Michael Vogt
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import print_function

import aptsources.distro
from datetime import datetime
import distro_info
from functools import wraps
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gio, Gtk

import logging
LOG=logging.getLogger(__name__)

import time

def setup_ui(self, path, domain):
    # setup ui
    self.builder = Gtk.Builder()
    self.builder.set_translation_domain(domain)
    self.builder.add_from_file(path)
    self.builder.connect_signals(self)
    for o in self.builder.get_objects():
        if issubclass(type(o), Gtk.Buildable):
            name = Gtk.Buildable.get_name(o)
            setattr(self, name, o)
        else:
            logging.debug("can not get name for object '%s'" % o)

def has_gnome_online_accounts():
    try:
        d = Gio.DesktopAppInfo.new('gnome-online-accounts-panel.desktop')
        return d != None
    except Exception:
        return False

def is_current_distro_lts():
    distro = aptsources.distro.get_distro()
    di = distro_info.UbuntuDistroInfo()
    return di.is_lts(distro.codename)

def is_current_distro_supported():
    distro = aptsources.distro.get_distro()
    di = distro_info.UbuntuDistroInfo()
    return distro.codename in di.supported(datetime.now().date())

def retry(exceptions, tries=10, delay=0.1, backoff=2):
    """
    Retry calling the decorated function using an exponential backoff.

    Args:
        exceptions: The exception to check. may be a tuple of
            exceptions to check.
        tries: Number of times to try (not retry) before giving up.
        delay: Initial delay between retries in seconds.
        backoff: Backoff multiplier (e.g. value of 2 will double the delay
            each retry).
    """
    def deco_retry(f):

        @wraps(f)
        def f_retry(*args, **kwargs):
            mtries, mdelay = tries, delay
            while mtries > 1:
                try:
                    return f(*args, **kwargs)
                except exceptions as e:
                    msg = '{}, Retrying in {} seconds...'.format(e, mdelay)
                    logging.warning(msg)
                    time.sleep(mdelay)
                    mtries -= 1
                    mdelay *= backoff
            return f(*args, **kwargs)

        return f_retry  # true decorator

    return deco_retry



#  GtkCdomProgress - add a cdrom to the apt sources
#
#  Copyright (c) 2004-2007 Canonical Ltd.
#                2004-2005 Michiel Sikkes
#
#  Author: Michael Vogt <mvo@debian.org>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA
import os

import apt
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk
from gettext import gettext as _

from softwareproperties.gtk.utils import (
  setup_ui,
)


class CdromProgress(apt.progress.base.CdromProgress):
  def __init__(self, datadir, parent):
    # gtk stuff
    setup_ui(self, os.path.join(datadir, "gtkbuilder", "dialog-cdrom-progress.ui"), domain="software-properties")
    
    self.dialog_cdrom_progress.show()
    self.dialog_cdrom_progress.set_transient_for(parent)
    self.parent = parent
    self.button_cdrom_close.set_sensitive(False)
    
  def update(self, text, step):
    """ update is called regularly so that the gui can be redrawn """
    if step > 0:
      self.progressbar_cdrom.set_fraction(step/float(self.totalSteps))
      if step == self.totalSteps:
        self.button_cdrom_close.set_sensitive(True)
    if text != "":
      self.label_cdrom.set_text(text)
    while Gtk.events_pending():
      Gtk.main_iteration()
  def askCdromName(self):
    dialog = Gtk.MessageDialog(parent=self.dialog_cdrom_progress,
                               flags=Gtk.DialogFlags.MODAL,
                               type=Gtk.MessageType.QUESTION,
                               buttons=Gtk.ButtonsType.OK_CANCEL,
                               message_format=None)
    dialog.set_markup(_("Please enter a name for the disc"))
    entry = Gtk.Entry()
    entry.show()
    dialog.vbox.pack_start(entry, True, True, 0)
    res = dialog.run()
    dialog.destroy()
    if res == Gtk.ResponseType.OK:
      name = entry.get_text()
      return (True,name)
    return (False,"")
  def changeCdrom(self):
    dialog = Gtk.MessageDialog(parent=self.dialog_cdrom_progress,
                               flags=Gtk.DialogFlags.MODAL,
                               type=Gtk.MessageType.QUESTION,
                               buttons=Gtk.ButtonsType.OK_CANCEL,
                               message_format=None)
    dialog.set_markup(_("Please insert a disk in the drive:"))
    dialog.set_position(Gtk.WindowPosition.CENTER_ON_PARENT)
    res = dialog.run()
    dialog.destroy()
    if res == Gtk.ResponseType.OK:
      return True
    return False



#
#  Copyright (c) 2019 Canonical
#
#  Authors:
#       Andrea Azzarone <andrea.azzarone@canonical.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

from gettext import gettext as _
import logging

import gi
from gi.repository import Gio, GLib, GObject

try:
    import dateutil.parser
    import requests_unixsocket

    gi.require_version('Snapd', '1')
    from gi.repository import Snapd
except(ImportError, ValueError):
    pass

from softwareproperties.gtk.utils import (
    has_gnome_online_accounts,
    is_current_distro_lts,
    is_current_distro_supported,
    retry
)

from softwareproperties.LivepatchSnap import LivepatchSnap


def datetime_parser(json_dict):
    for (key, value) in json_dict.items():
        try:
            json_dict[key] = dateutil.parser.parse(value)
        except (ValueError, TypeError):
            pass
    return json_dict

class LivepatchAvailability:
    FALSE = 0
    TRUE = 1
    NO_CONNECTIVITY=3
    CHECKING = 2


class LivepatchService(GObject.GObject):

    # Constants
    STATUS_ENDPOINT = 'http+unix://%2Fvar%2Fsnap%2Fcanonical-livepatch%2Fcurrent%2Flivepatchd.sock/status'
    ENABLE_ENDPOINT = 'http+unix://%2Fvar%2Fsnap%2Fcanonical-livepatch%2Fcurrent%2Flivepatchd-priv.sock/enable'
    DISABLE_ENDPOINT = 'http+unix://%2Fvar%2Fsnap%2Fcanonical-livepatch%2Fcurrent%2Flivepatchd-priv.sock/disable'
    LIVEPATCH_RUNNING_FILE = '/var/snap/canonical-livepatch/common/machine-token'

    ENABLE_ERROR_MSG = _('Failed to enable Livepatch: {}')
    DISABLE_ERROR_MSG = _('Failed to disable Livepatch: {}')

    # GObject.GObject
    __gproperties__ = {
        'availability': (
            int, None, None,
            LivepatchAvailability.FALSE,
            LivepatchAvailability.CHECKING,
            LivepatchAvailability.FALSE,
            GObject.PARAM_READABLE),
        'availability-message': (
            str, None, None, None, GObject.PARAM_READABLE),
        'enabled': (
            bool, None, None, False, GObject.PARAM_READABLE),
    }

    def __init__(self):
        GObject.GObject.__init__(self)

        self._timeout_id = 0

        self._snap = LivepatchSnap()
        self._session = requests_unixsocket.Session()

        # Init Properties
        self._availability = LivepatchAvailability.FALSE
        self._availability_message = None
        lp_file = Gio.File.new_for_path(path=self.LIVEPATCH_RUNNING_FILE)
        self._enabled = lp_file.query_exists()

        # Monitor connectivity status
        self._nm = Gio.NetworkMonitor.get_default()
        self._nm.connect('notify::connectivity', self._network_changed_cb)

        # Monitor status of canonical-livepatch
        self._lp_monitor = lp_file.monitor_file(Gio.FileMonitorFlags.NONE)
        self._lp_monitor.connect('changed', self._livepatch_enabled_changed_cb)

    def do_get_property(self, pspec):
        if pspec.name == 'availability':
            return self._availability
        elif pspec.name == 'availability-message':
            return self._availability_message
        elif pspec.name == 'enabled':
            return self._enabled
        else:
            raise AssertionError

    # Public API
    def trigger_availability_check(self):
        """Trigger a Livepatch availability check to be executed after a short
        timeout. Multiple triggers will result in a single request.

        A notify::availability will be emitted when the check starts, and
        another one when the check ends.
        """
        def _update_availability():
            # each rule is a tuple of two elements, a callable and a string. The
            # string rapresents the error message that needs to be shown if the
            # callable returns false.
            rules = [
                (lambda: self._snap.get_status() != Snapd.SnapStatus.UNKNOWN,
                    _('Canonical Livepatch snap is not available.')),
                (has_gnome_online_accounts,
                    _('Gnome Online Accounts is required to enable Livepatch.')),
                (is_current_distro_lts,
                    _('Livepatch is not available for this release.')),
                (is_current_distro_supported,
                    _('The current release is no longer supported.'))]

            if self._nm.props.connectivity != Gio.NetworkConnectivity.FULL:
                self._availability = LivepatchAvailability.NO_CONNECTIVITY
                self._availability_message = None
            else:
                for func, message in rules:
                    if not func():
                        self._availability = LivepatchAvailability.FALSE
                        self._availability_message = message
                        break
                else:
                    self._availability = LivepatchAvailability.TRUE
                    self._availability_message = None

            self.notify('availability')
            self.notify('availability-message')

            self._timeout_id = 0
            return False

        self._availability = LivepatchAvailability.CHECKING
        self._availability_message = None
        self.notify('availability')
        self.notify('availability-message')

        if self._timeout_id == 0:
            self._timeout_id = GLib.timeout_add_seconds(3, _update_availability)

    def set_enabled(self, enabled, token):
        """Enable or disable Canonical Livepatch in the current system. This
        function will return once the operation succeeded or failed.

        Args:
            enabled(bool): wheater to enable or disable the service.
            token(str): the authentication token to be used to enable Canonical
                Livepatch service.

        Returns:
            (False, '') if successful, (True, error_message) otherwise.
        """
        if self._enabled == enabled:
            return False, ''

        if not enabled:
            return self._disable_service()
        elif self._snap.get_status() == Snapd.SnapStatus.ACTIVE:
            return self._enable_service(token)
        else:
            success, msg = self._snap.enable_or_install()
            return self._enable_service(token) if success else (True, msg)

    def get_status(self):
        """Synchronously retrieve the status of Canonical Livepatch.

        Returns:
            str: The status. A valid string for success, None otherwise.
        """
        try:
            params = {'verbosity': 3, 'format': 'json'}
            r = self._session.get(self.STATUS_ENDPOINT, params=params)
            return r.json(object_hook=datetime_parser)
        except Exception as e:
            logging.debug('Failed to get Livepatch status: {}'.format(str(e)))
            return None

    # Private methods
    def _enable_service(self, token):
        """Enable Canonical Livepatch in the current system. This function will
        return once the operation succeeded or failed.

        Args:
            token(str): the authentication token to be used to enable Canonical
                Livepatch service.

        Returns:
            (False, '') if successful, (True, error_message) otherwise.
        """
        try:
            return self._enable_service_with_retry(token)
        except Exception as e:
            return True, self.ENABLE_ERROR_MSG.format(str(e))

    @retry(Exception)
    def _enable_service_with_retry(self, token):
        params = {'auth-token': token}
        r = self._session.put(self.ENABLE_ENDPOINT, params=params)
        return not r.ok, '' if r.ok else self.ENABLE_ERROR_MSG.format(r.text)

    def _disable_service(self):
        """Disable Canonical Livepatch in the current system. This function will
        return once the operation succeeded or failed.

        Returns:
            (False, '') if successful, (True, error_message) otherwise.
        """
        try:
            return self._disable_service_with_retry()
        except Exception as e:
            return True, self.DISABLE_ERROR_MSG.format(str(e))


    @retry(Exception)
    def _disable_service_with_retry(self):
        r = self._session.put(self.DISABLE_ENDPOINT)
        return not r.ok, '' if r.ok else self.DISABLE_ERROR_MSG.format(r.text)

    # Signals handlers
    def _network_changed_cb(self, monitor, network_available):
        self.trigger_availability_check()

    def _livepatch_enabled_changed_cb(self, fm, file, other_file, event_type):
        enabled = file.query_exists()
        if self._enabled != enabled:
            self._enabled = enabled
            self.notify('enabled')



#!/usr/bin/env python3

import setuptools
setuptools  # pyflakes
from DistUtilsExtra.auto import setup


setup(name='software-properties',
      version='0.96',
      packages=[
                'softwareproperties',
                'softwareproperties.dbus',
                'softwareproperties.gtk',
                'softwareproperties.qt',
                ],
      scripts=[
               'software-properties-gtk',
               'software-properties-qt',
               'add-apt-repository',
               ],
      data_files=[
                  ('lib/software-properties/',
                   ['software-properties-dbus', ]
                  ),
                  ],
      test_suite="tests",
     )



# dialog_edit.py.in - edit a existing repository
#  
#  Copyright (c) 2004-2005 Canonical
#                2005 Michiel Sikkes
#  
#  Authors: 
#       Michael Vogt <mvo@debian.org>
#       Michiel Sikkes <michiels@gnome.org>
# 
#  This program is free software; you can redistribute it and/or 
#  modify it under the terms of the GNU General Public License as 
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
# 
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
# 
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

from __future__ import print_function

import os
import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

from aptsources.sourceslist import SourceEntry
from softwareproperties.gtk.utils import (
    setup_ui,
)

class DialogEdit:
  def __init__(self, parent, sourceslist, source_entry, datadir):
    self.sourceslist = sourceslist
    self.source_entry = source_entry

    # gtk stuff
    setup_ui(self, os.path.join(datadir, "gtkbuilder", "dialog-edit-source.ui"), domain="software-properties")
    self.main = self.dialog_edit
    self.main.set_transient_for(parent)
    
    # type
    if source_entry.type == "deb":
      self.combobox_type.set_active(0)
    elif source_entry.type == "deb-src":
      self.combobox_type.set_active(1)
    else:
      print("Error, unknown source type: '%s'" % source_entry.type)

    # uri
    self.entry_uri.set_text(source_entry.uri)
    self.entry_dist.set_text(source_entry.dist)

    comps = ""
    for c in source_entry.comps:
      if len(comps) > 0:
        comps = comps + " " + c
      else:
        comps = c
    self.entry_comps.set_text(comps)

    self.entry_comment.set_text(source_entry.comment)

    # finally set the signal so that the check function is not tiggered 
    # during initialisation
    for entry in (self.entry_uri, self.entry_dist, self.entry_comps, self.entry_comment):
        entry.connect("changed", self.check_line)

  def check_line(self, *args):
    """Check for a valid apt line and set the sensitiveness of the
       button 'add' accordingly"""
    line = self.get_line()
    if line == False:
      self.button_edit_ok.set_sensitive(False)
      return
    source_entry = SourceEntry(line)
    if source_entry.invalid == True:
      self.button_edit_ok.set_sensitive(False)
    else:
      self.button_edit_ok.set_sensitive(True)

  def get_line(self):
    """Collect all values from the entries and create an apt line"""
    if self.source_entry.disabled == True:
      line = "#"
    else:
      line = ""

    if self.combobox_type.get_active() == 0:
      line = "%sdeb" % line
    else:
      line = "%sdeb-src" % line

    text = self.entry_uri.get_text()
    if len(text) < 1 or text.find(" ") != -1 or text.find("#") != -1:
      return False  
    line = line + " " + self.entry_uri.get_text()

    text = self.entry_dist.get_text()
    if len(text) < 1 or text.find(" ") != -1 or text.find("#") != -1:
      return False    
    line = line + " " + self.entry_dist.get_text()

    text = self.entry_comps.get_text()
    if text.find("#") != -1:
      return False    
    elif text != "":
      line = line + " " + self.entry_comps.get_text()

    if self.entry_comment.get_text() != "":
      line = line + " #" + self.entry_comment.get_text() + "\n"
    else:
      line = line + "\n"
    return line
          
  def run(self):
      res = self.main.run()
      if res == Gtk.ResponseType.OK:
        line = self.get_line()

        # change repository
        index = self.sourceslist.list.index(self.source_entry)
        file = self.sourceslist.list[index].file
        self.new_source_entry = SourceEntry(line,file)
        self.sourceslist.list[index] = self.new_source_entry
      self.main.hide()
      return res



#  i18n reusable helper class
#
#  Copyright (c) 2009 Canonical Ltd.
#
#  Author: Amichai Rothman <amichai2@amichais.net>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

#
# This is a reusable helper component containing various i18n related
# utilities. It also helps narrow the gap between GTK and Qt frontends
# to a given application, by providing the translate_widget function to
# translate both language and GTK accelerators and styles to Qt widgets.
# I hope to see this improve and grow into an abstract dual-gui helper
# framework, to make it much easier to port Ubuntu/GTK apps to Kubuntu
# as well as maintain them as they grow.
#

from gettext import gettext as _
import re

# FIXME: what's the point of these functions????
from PyQt5.QtCore import *
from PyQt5.QtGui import *
from PyQt5.QtWidgets import *

def strip_html(str):
  regex = re.compile('<[^>]*>')
  return regex.sub('', str)

def translate_string(text, nohtml=False, accelerators='replace'):
  if not text:
    return text
  translated = _(str(text)) # text might be QString, unicode or str
  if nohtml:
    translated = strip_html(translated)
  if accelerators == 'replace':
    translated = translated.replace('_', '&') # convert accelerators
  elif accelerators == 'remove':
    translated = translated.replace('_', '') # convert accelerators
  return translated

def apply_styles(widget, str):
  # this is obviously just a start :-)
  if str.startswith("<b>"):
      widget.setStyleSheet("QGroupBox {font-weight:bold;}");

def translate_widget(widget, recursive=True):
  """ Translate a widget recursively, while converting GTK
      accelerators and styles to appropriate QT equivalents.
      The given widget is typically a dialog or main window,
      translated after its components are initialized but
      before being shown.
      """
  if isinstance(widget, QTabWidget):
    for i in range(0, widget.count()):
      widget.setTabText(i, translate_string(widget.tabText(i)))
  elif isinstance(widget, QGroupBox):
    apply_styles(widget, widget.title())
    widget.setTitle(translate_string(widget.title(), nohtml=True))
  elif isinstance(widget, QLineEdit):
    widget.setText(widget.text()) # user-editable text: utf8 without translation
  elif isinstance(widget, QWidget):
    widget.setWindowTitle(translate_string(widget.windowTitle(), accelerators='remove'))
    try:
      widget.setText(translate_string(widget.text()))
    except AttributeError:
      pass

  if recursive:
    for child in widget.children():
      translate_widget(child)



#!/usr/bin/python

import apt

import unittest
import sys
try:
    from urllib.request import urlopen
    from urllib.error import HTTPError, URLError
except ImportError:
    from urllib2 import HTTPError, URLError, urlopen
try:
    from http.client import HTTPException
except ImportError:
    from httplib import HTTPException

sys.path.insert(0, "..")

from softwareproperties.SoftwareProperties import shortcut_handler
from softwareproperties.shortcuts import ShortcutException
from mock import patch

def has_network():
    try:
        network = urlopen("https://launchpad.net/")
        network
    except (URLError, HTTPException):
        return False
    return True

class ShortcutsTestcase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        for k in apt.apt_pkg.config.keys():
            apt.apt_pkg.config.clear(k)
        apt.apt_pkg.init()

    def setUp(self):
        # avoid polution from other tests
        apt.apt_pkg.config.set("Dir", "/")
        apt.apt_pkg.config.set("Dir::Etc", "etc/apt")
        apt.apt_pkg.config.set("Dir::Etc::sourcelist", "sources.list")
        apt.apt_pkg.config.set("Dir::Etc::sourceparts", "sources.list.d")

    def test_shortcut_none(self):
        line = "deb http://ubuntu.com/ubuntu trusty main"
        handler = shortcut_handler(line)
        self.assertEqual((line, None), handler.expand())

    @unittest.skipUnless(has_network(), "requires network")
    def test_shortcut_ppa(self):
        line = "ppa:mvo"
        handler = shortcut_handler(line)
        self.assertEqual(
            ('deb http://ppa.launchpad.net/mvo/ppa/ubuntu trusty main',
             '/etc/apt/sources.list.d/mvo-ubuntu-ppa-trusty.list'),
            handler.expand("trusty", distro="ubuntu"))

    @unittest.skipUnless(has_network(), "requires network")
    def test_shortcut_cloudarchive(self):
        line = "cloud-archive:folsom"
        handler = shortcut_handler(line)
        self.assertEqual(
            ('deb http://ubuntu-cloud.archive.canonical.com/ubuntu '\
             'precise-updates/folsom main',
             '/etc/apt/sources.list.d/cloudarchive-folsom.list'),
            handler.expand("precise", distro="ubuntu"))

    def test_shortcut_exception(self):
        with self.assertRaises(ShortcutException):
            with patch('softwareproperties.ppa.get_ppa_info_from_lp',
                       side_effect=lambda *args: HTTPError("url", 404, "not found", [], None)):
                shortcut_handler("ppa:mvo")



if __name__ == "__main__":
    unittest.main()



#
#  Copyright (c) 2018 Canonical
#
#  Authors:
#       Andrea Azzarone <andrea.azzarone@canonical.com>
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License as
#  published by the Free Software Foundation; either version 2 of the
#  License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
#  USA

import gi
gi.require_version('Goa', '1.0')
from gi.repository import Gio, GLib, Goa, GObject

import logging

class GoaAuth(GObject.GObject):

    # Properties
    logged = GObject.Property(type=bool, default=False)
    username = GObject.Property(type=str, default=None)

    def __init__(self):
        GObject.GObject.__init__(self)

        self.account = None
        self.cancellable = Gio.Cancellable()
        Goa.Client.new(self.cancellable, self._on_goa_client_ready)

        self.settings = Gio.Settings.new('com.ubuntu.SoftwareProperties')
        self.settings.connect('changed::goa-account-id', self._on_settings_changed)

    def _on_goa_client_ready(self, source, res):
        try:
            self.goa_client = Goa.Client.new_finish(res)
        except GLib.Error as e:
            logging.error('Failed to get a Gnome Online Account: {}'.format(e.message))
            self.goa_client = None
        else:
            self._load()

    def login(self, account):
        assert(account)
        self._update_state(account)
        self._store()

    def logout(self):
        self._update_state(None)
        self._store()

    @GObject.Property
    def token(self):
        if self.account is None or self.goa_client is None:
            return None

        obj = self.goa_client.lookup_by_id(self.account.props.id)
        if obj is None:
            return None

        pbased = obj.get_password_based()
        if pbased is None:
            return None

        return pbased.call_get_password_sync('livepatch')

    def _update_state_from_account_id(self, account_id):
        if account_id and self.goa_client is not None:
            # Make sure the account-id is valid
            obj = self.goa_client.lookup_by_id(account_id)
            if obj is None:
                self._update_state(None)
                return

            account = obj.get_account()
            if account is None:
                self._update_state(None)
                return

            self._update_state(account)
        else:
            self._update_state(None)

    def _update_state(self, account):
        self.account = account
        if self.account is None:
            self.username = None
            self.logged = False
        else:
            try:
                account.call_ensure_credentials_sync(None)
            except Exception:
                self.username = None
                self.logged = False
            else:
                self.account.connect('notify::attention-needed', lambda o, v: self.logout())
                self.username = self.account.props.presentation_identity
                self.logged = True

    def _on_settings_changed(self, settings, key):
        self._load()

    def _load(self):
        # Retrieve the stored account-id
        account_id = self.settings.get_string('goa-account-id')
        self._update_state_from_account_id(account_id)

    def _store(self):
        # Store the account-id
        if self.logged:
            account_id = self.account.props.id
            self.settings.set_string('goa-account-id', account_id)
        else:
            self.settings.set_string('goa-account-id', "")




